"""The HighChart pane enables you to use the Highchart chart in Panel """
from typing import Dict

import param
from panel.pane.base import PaneBase
from panel.util import lazy_load, string_types
from pyviz_comms import JupyterComm

from . import config  # pylint: disable=unused-import


class HighChartBase(PaneBase):
    """A Base Panel HighChart pane"""

    # Events: https://api.highcharts.com/highcharts/plotOptions.series.point.events
    object = param.Dict(
        default=None,
        doc="""
        The initial user configuration of the chart. Will reset the chart when ever a
        configuration is provided.""",
    )
    object_update = param.Dict(doc="""Incremental change to the initial user configuration.""")

    event = param.Dict(doc="""Events raised by the chart""", readonly=True)

    _add_series = param.Dict(doc="""Transfer the arguments of the add_series method""")

    priority = None

    _rename = {"object": "config", "object_update": "config_update"}

    _updates = True

    _model_module = "panel_highcharts.models.highchart"
    _model = "HighChart"

    @classmethod
    def applies(cls, obj):
        if isinstance(obj, (dict, string_types)):
            return 0
        return False

    def _get_model(self, doc, root=None, parent=None, comm=None):
        _BkHighChart = lazy_load(  # pylint: disable=invalid-name
            self._model_module, self._model, isinstance(comm, JupyterComm)
        )
        props = self._process_param_change(self._init_params())
        if not "config_update" in props:
            props["config_update"] = self.object_update
        model = _BkHighChart(config=self.object, **props)
        root = root or model
        self._link_props(
            model,
            [
                "event",
            ],
            doc,
            root,
            comm,
        )
        self._models[root.ref["id"]] = (model, parent)
        return model

    def _update(self, ref=None, model=None):
        model.update(config=self.object, config_update=self.object_update)

    def add_series(self, options: Dict, redraw: bool = True, animation: bool = True):
        """Adds the seried specified by the options

        C.f. https://api.highcharts.com/class-reference/Highcharts.Chart#addSeries

        Args:
            options (Dict): The config options for the series.
            redraw (bool, optional): Whether to redraw the chart after adding. Defaults to True.
            animation (bool, optional): Whether to apply animation. Defaults to True.
        """
        self._add_series = {
            "options": options,
            "redraw": redraw,
            "animation": animation,
        }


class HighChart(HighChartBase):
    """A Panel HighChart pane"""

    _model_module = "panel_highcharts.models.highchart"
    _model = "HighChart"


class HighStock(HighChartBase):
    """A Panel HighStock pane"""

    _model_module = "panel_highcharts.models.highstock"
    _model = "HighStock"


class HighMap(HighChartBase):
    """A Panel HighMap pane"""

    _model_module = "panel_highcharts.models.highmap"
    _model = "HighMap"


class HighGantt(HighChartBase):
    """A Panel HighGantt pane"""

    _model_module = "panel_highcharts.models.highgantt"
    _model = "HighGantt"



"""Use

- panel_highcharts.config.js_files to define which Highcharts js files to include
- the HighChart pane to use the Highcharts Chart in Panel.
"""
from . import config
from .charts import HighChart, HighGantt, HighMap, HighStock



"""Contains the Bokeh Model for the HighMap pane"""
from collections import OrderedDict
from typing import List, Optional

from .highbase import PATH_HIGH_CHARTS, PATH_REQUIRE_HIGH_CHARTS, PATHS, HighBase


class HighMap(HighBase):
    """The Bokeh Model for the HighMap pane"""

    __javascript__ = [
        "https://code.highcharts.com/highcharts.js",
        "https://code.highcharts.com/maps/modules/map.js",
        "https://code.highcharts.com/modules/export-data.js",
        "https://code.highcharts.com/modules/exporting.js",
    ]

    # https://api.highcharts.com/class-reference/#toc5
    __js_require__ = {
        "packages": [
            {
                "name": "highcharts",
                "main": "highcharts",
            },
        ],
        "paths": {
            "highcharts": "https://code.highcharts.com",
            "highcharts/modules/map": "https://code.highcharts.com/maps/modules/map",
            "highcharts/modules/exporting": "https://code.highcharts.com/modules/exporting",
            "highcharts/modules/export-data": "https://code.highcharts.com/modules/export-data",
        },
        "exports": {
            "highcharts": "Highcharts",
            "highcharts/modules/map": "highchartsmodulesmap",
            "highcharts/modules/exporting": "highchartsmodulesexporting",
            "highcharts/modules/export-data": "highchartsmodulesexportdata",
        },
    }

    @classmethod
    def js_files(  # pylint: disable=too-many-locals, too-many-arguments
        cls,
        highcharts_accessibility: bool = False,
        highcharts_annotations: bool = False,
        highcharts_boost: bool = False,
        highcharts_broken_axis: bool = False,
        highcharts_canvas_tools: bool = False,
        highcharts_coloraxis: bool = False,
        highcharts_data: bool = False,
        highcharts_drilldown: bool = False,
        highcharts_export_data: bool = True,
        highcharts_exporting: bool = True,
        highcharts_marker_clusters: bool = False,
        highcharts_more: bool = False,
        highcharts_networkgraph: bool = False,
        highcharts_no_data: bool = False,
        highcharts_offline_exporting: bool = False,
        mapdata: Optional[List[str]] = None,
    ):
        """Configures the js files to include from https://code.highcharts.com

        Use this before using `panel.extension("highchart")`

        Args:
            highcharts_accessibility (bool, optional): Defaults to False.
            highcharts_annotations (bool, optional): Defaults to False.
            highcharts_boost (bool, optional): Defaults to False.
            highcharts_broken_axis (bool, optional): Defaults to False.
            highcharts_canvas_tools (bool, optional): Defaults to False.
            highcharts_coloraxis (bool, optional): Defaults to False.
            highcharts_data (bool, optional): Defaults to False.
            highcharts_drilldown (bool, optional): Defaults to False.
            highcharts_export_data (bool, optional): Defaults to True.
            highcharts_exporting (bool, optional): Defaults to True.
            highcharts_marker_clusters (bool, optional): Defaults to False.
            highcharts_more (bool, optional): Defaults to False.
            highcharts_networkgraph (bool, optional): Defaults to False.
            highcharts_no_data (bool, optional): Defaults to False.
            highcharts_offline_exporting (bool, optional): Defaults to False.
        """
        paths = OrderedDict()
        include = {
            "highcharts/modules/map": True,
            "highcharts-more": highcharts_more,
            "highcharts/modules/accessibility": highcharts_accessibility,
            "highcharts/modules/annotations": highcharts_annotations,
            "highcharts/modules/boost": highcharts_boost,
            "highcharts/modules/broken-axis": highcharts_broken_axis,
            "highcharts/modules/canvas-tools": highcharts_canvas_tools,
            "highcharts/modules/coloraxis": highcharts_coloraxis,
            "highcharts/modules/data": highcharts_data,
            "highcharts/modules/drilldown": highcharts_drilldown,
            "highcharts/modules/export-data": highcharts_export_data,
            "highcharts/modules/exporting": highcharts_exporting,
            "highcharts/modules/marker-clusters": highcharts_marker_clusters,
            "highcharts/modules/networkgraph": highcharts_networkgraph,
            "highcharts/modules/no-data": highcharts_no_data,
            "highcharts/modules/offline-exporting": highcharts_offline_exporting,
        }
        for key, value in include.items():
            if value:
                paths[key] = PATHS[key]

        if mapdata:
            for key in mapdata:
                paths[f"mapdata/{key}"] = f"{PATH_REQUIRE_HIGH_CHARTS}/mapdata/{key}.js"

        cls.__javascript__ = [PATH_HIGH_CHARTS] + list(paths.values())

        cls.__js_require__["paths"] = {
            "highcharts": PATH_REQUIRE_HIGH_CHARTS,
            **{k: v[:-3] for k, v in paths.items()},
        }
        cls.__js_require__["exports"] = {
            "highcharts": "Highcharts",
            **{k: k.replace("/", "").replace("-", "") for k in paths},  # type: ignore
        }



"""Contains the Bokeh Model for the HighBase pane"""
from collections import OrderedDict

from bokeh.core.properties import Any, Dict, Nullable, String
from bokeh.models import LayoutDOM
from panel.util import classproperty

PATH_HIGH_CHARTS = "https://code.highcharts.com/highcharts.js"
PATH_REQUIRE_HIGH_CHARTS = "https://code.highcharts.com"

PATHS = OrderedDict(
    [
        ("highcharts-3d", "https://code.highcharts.com/highcharts-3d.js"),
        ("highcharts-more", "https://code.highcharts.com/highcharts-more.js"),
        ("highcharts/modules/stock", "https://code.highcharts.com/modules/stock.js"),
        ("highcharts/modules/map", "https://code.highcharts.com/maps/modules/map.js"),
        ("highcharts/modules/coloraxis", "https://code.highcharts.com/maps/modules/coloraxis.js"),
        (
            "highcharts/modules/marker-clusters",
            "https://code.highcharts.com/maps/modules/marker-clusters.js",
        ),
        ("highcharts/modules/gantt", "https://code.highcharts.com/modules/gantt.js"),
        ("highcharts/modules/exporting", "https://code.highcharts.com/modules/exporting.js"),
        ("highcharts/modules/export-data", "https://code.highcharts.com/modules/export-data.js"),
        ("highcharts/modules/networkgraph", "https://code.highcharts.com/modules/networkgraph.js"),
        ("highcharts/modules/annotations", "https://code.highcharts.com/modules/annotations.js"),
        ("highcharts/modules/boost", "https://code.highcharts.com/modules/boost.js"),
        ("highcharts/modules/broken-axis", "https://code.highcharts.com/modules/broken-axis.js"),
        ("highcharts/modules/canvas-tools", "https://code.highcharts.com/modules/canvas-tools.js"),
        ("highcharts/modules/data", "https://code.highcharts.com/modules/data.js"),
        ("highcharts/modules/drilldown", "https://code.highcharts.com/modules/drilldown.js"),
        ("highcharts/modules/funnel", "https://code.highcharts.com/modules/funnel.js"),
        ("highcharts/modules/heatmap", "https://code.highcharts.com/modules/heatmap.js"),
        (
            "highcharts/modules/no-data",
            "https://code.highcharts.com/modules/no-data-to-display.js",
        ),
        (
            "highcharts/modules/offline-exporting",
            "https://code.highcharts.com/modules/offline-exporting.js",
        ),
        ("highcharts/modules/solid-gauge", "https://code.highcharts.com/modules/solid-gauge.js"),
        ("highcharts/modules/treemap", "https://code.highcharts.com/modules/treemap.js"),
        ("highcharts/modules/variwide", "https://code.highcharts.com/modules/variwide.js"),
    ]
)


class HighBase(LayoutDOM):
    """The Bokeh Model for the HighBase pane"""

    config = Nullable(Dict(String, Any))
    config_update = Nullable(Dict(String, Any))
    _add_series = Nullable(Dict(String, Any))

    # Events
    event = Nullable(Dict(String, Any))

    __javascript__ = [
        "https://code.highcharts.com/highcharts.js",
        "https://code.highcharts.com/modules/export-data.js",
        "https://code.highcharts.com/modules/exporting.js",
    ]

    @classproperty
    def __js_skip__(cls) -> Dict:  # pylint: disable=no-self-argument
        return {
            "highcharts": cls.__javascript__,
        }

    # https://api.highcharts.com/class-reference/#toc5
    __js_require__ = {
        "packages": [
            {
                "name": "highcharts",
                "main": "highcharts",
            },
        ],
        "paths": {
            "highcharts": "https://code.highcharts.com",
            "highcharts/modules/exporting": "https://code.highcharts.com/modules/exporting",
            "highcharts/modules/export-data": "https://code.highcharts.com/modules/export-data",
        },
        "exports": {
            "highcharts": "Highcharts",
            "highcharts/modules/exporting": "highchartsmodulesexporting",
            "highcharts/modules/export-data": "highchartsmodulesexportdata",
        },
    }



"""Use

- panel_highcharts.config.js_files to define which Highcharts js files to include
"""
from typing import List, Optional

from panel import extension
from panel.io import notebook

# Enables using pn.extension("highchart", ...)
# pylint: disable=protected-access
extension._imports["highchart"] = "panel_highcharts.models.highchart"
extension._imports["highstock"] = "panel_highcharts.models.highstock"
extension._imports["highmap"] = "panel_highcharts.models.highmap"
extension._imports["highgantt"] = "panel_highcharts.models.highgantt"
# pylint: enable=protected-access


# pylint: disable=too-many-arguments, invalid-name
def _mock():
    """Temporary fix #2571 https://github.com/holoviz/panel/issues/2571"""

    def _autoload_js(
        bundle, configs, requirements, exports, skip_imports, ipywidget, load_timeout=5000
    ):
        config = {"packages": {}, "paths": {}, "shim": {}}
        for conf in configs:
            for key, c in conf.items():
                config[key].update(c)
        return notebook.AUTOLOAD_NB_JS.render(
            bundle=bundle,
            force=True,
            timeout=load_timeout,
            config=config,
            requirements=requirements,
            exports=exports,
            skip_imports=skip_imports,
            ipywidget=ipywidget,
        )

    notebook._autoload_js = _autoload_js  # pylint: disable=protected-access


# pylint: enable=too-many-arguments, invalid-name

_mock()


def js_files(  # pylint: disable=too-many-locals, too-many-arguments
    highcharts_accessibility: bool = False,
    highcharts_annotations: bool = False,
    highcharts_boost: bool = False,
    highcharts_broken_axis: bool = False,
    highcharts_canvas_tools: bool = False,
    highcharts_coloraxis: bool = False,
    highcharts_data: bool = False,
    highcharts_drilldown: bool = False,
    highcharts_export_data: bool = True,
    highcharts_exporting: bool = True,
    highcharts_funnel: bool = False,
    highcharts_heatmap: bool = False,
    highcharts_marker_clusters: bool = False,
    highcharts_more: bool = False,
    highcharts_networkgraph: bool = False,
    highcharts_no_data: bool = False,
    highcharts_offline_exporting: bool = False,
    highcharts_solid_gauge: bool = False,
    highcharts_3d: bool = False,
    highcharts_treemap: bool = False,
    highcharts_variwide: bool = False,
    mapdata: Optional[List[str]] = None,
):
    """Configures the js files to include from https://code.highcharts.com

    Use this before using `panel.extension("highchart")`

    Args:
        highcharts_accessibility (bool, optional): Defaults to False.
        highcharts_annotations (bool, optional): Defaults to False.
        highcharts_boost (bool, optional): Defaults to False.
        highcharts_broken_axis (bool, optional): Defaults to False.
        highcharts_coloraxis (bool, optional): Defaults to False.
        highcharts_canvas_tools (bool, optional): Defaults to False.
        highcharts_data (bool, optional): Defaults to False.
        highcharts_drilldown (bool, optional): Defaults to False.
        highcharts_export_data (bool, optional): Defaults to True.
        highcharts_exporting (bool, optional): Defaults to True.
        highcharts_funnel (bool, optional): Defaults to False.
        highcharts_heatmap (bool, optional): Defaults to False.
        highcharts_marker_clusters (bool, optional): Defaults to False.
        highcharts_more (bool, optional): Defaults to False.
        highcharts_networkgraph (bool, optional): Defaults to False.
        highcharts_no_data (bool, optional): Defaults to False.
        highcharts_offline_exporting (bool, optional): Defaults to False.
        highcharts_solid_gauge (bool, optional): Defaults to False.
        highcharts_3d (bool, optional): Defaults to False.
        highcharts_treemap (bool, optional): Defaults to False.
        highcharts_variwide (bool, optional): Defaults to False.
    """
    # pylint: disable=import-outside-toplevel
    from .models.highchart import HighChart
    from .models.highgantt import HighGantt
    from .models.highmap import HighMap
    from .models.highstock import HighStock

    HighChart.js_files(
        highcharts_accessibility=highcharts_accessibility,
        highcharts_annotations=highcharts_annotations,
        highcharts_boost=highcharts_boost,
        highcharts_broken_axis=highcharts_broken_axis,
        highcharts_canvas_tools=highcharts_canvas_tools,
        highcharts_data=highcharts_data,
        highcharts_drilldown=highcharts_drilldown,
        highcharts_export_data=highcharts_export_data,
        highcharts_exporting=highcharts_exporting,
        highcharts_funnel=highcharts_funnel,
        highcharts_heatmap=highcharts_heatmap,
        highcharts_more=highcharts_more,
        highcharts_networkgraph=highcharts_networkgraph,
        highcharts_no_data=highcharts_no_data,
        highcharts_offline_exporting=highcharts_offline_exporting,
        highcharts_solid_gauge=highcharts_solid_gauge,
        highcharts_3d=highcharts_3d,
        highcharts_treemap=highcharts_treemap,
        highcharts_variwide=highcharts_variwide,
    )

    HighStock.js_files(
        highcharts_accessibility=highcharts_accessibility,
        highcharts_annotations=highcharts_annotations,
        highcharts_boost=highcharts_boost,
        highcharts_broken_axis=highcharts_broken_axis,
        highcharts_canvas_tools=highcharts_canvas_tools,
        highcharts_data=highcharts_data,
        highcharts_drilldown=highcharts_drilldown,
        highcharts_export_data=highcharts_export_data,
        highcharts_exporting=highcharts_exporting,
        highcharts_more=highcharts_more,
        highcharts_no_data=highcharts_no_data,
        highcharts_offline_exporting=highcharts_offline_exporting,
        highcharts_solid_gauge=highcharts_solid_gauge,
    )

    HighMap.js_files(
        highcharts_accessibility=highcharts_accessibility,
        highcharts_annotations=highcharts_annotations,
        highcharts_boost=highcharts_boost,
        highcharts_broken_axis=highcharts_broken_axis,
        highcharts_canvas_tools=highcharts_canvas_tools,
        highcharts_coloraxis=highcharts_coloraxis,
        highcharts_data=highcharts_data,
        highcharts_drilldown=highcharts_drilldown,
        highcharts_export_data=highcharts_export_data,
        highcharts_exporting=highcharts_exporting,
        highcharts_marker_clusters=highcharts_marker_clusters,
        highcharts_more=highcharts_more,
        highcharts_no_data=highcharts_no_data,
        highcharts_offline_exporting=highcharts_offline_exporting,
        mapdata=mapdata,
    )

    HighGantt.js_files(
        highcharts_accessibility=highcharts_accessibility,
        highcharts_annotations=highcharts_annotations,
        highcharts_boost=highcharts_boost,
        highcharts_broken_axis=highcharts_broken_axis,
        highcharts_canvas_tools=highcharts_canvas_tools,
        highcharts_data=highcharts_data,
        highcharts_drilldown=highcharts_drilldown,
        highcharts_export_data=highcharts_export_data,
        highcharts_exporting=highcharts_exporting,
        highcharts_more=highcharts_more,
        highcharts_no_data=highcharts_no_data,
        highcharts_offline_exporting=highcharts_offline_exporting,
    )
